use core::mem;

use cortex_m::prelude::_embedded_hal_spi_FullDuplex;
use embedded_hal::digital::v2::OutputPin;
use fugit::HertzU32;
use rp_pico::{
    hal::{
        gpio::{
            self,
            bank0::{Gpio2, Gpio3, Gpio4, Gpio5},
            Function, Output, Pin, PushPull, Spi,
        },
        spi::{self, Enabled, Spi as SpiDevice},
    },
    pac::{RESETS, SPI0},
    Pins,
};

pub struct MouseSensor {
    _sclk: Pin<Gpio2, Function<Spi>>,
    _mosi: Pin<Gpio3, Function<Spi>>,
    _miso: Pin<Gpio4, Function<Spi>>,
    cs: Pin<Gpio5, Output<PushPull>>,

    spi: SpiDevice<Enabled, SPI0, 8>,
}

impl MouseSensor {
    const READ: u8 = 0x7F;
    const WRITE: u8 = 0x80;

    const REG_PRODUCT_ID1: u8 = 0x00;
    const REG_PRODUCT_ID2: u8 = 0x01;
    const REG_MOTION_STATUS: u8 = 0x02;
    const REG_DELTA_X: u8 = 0x03;
    const REG_DELTA_Y: u8 = 0x04;
    const REG_OPERATION_MODE: u8 = 0x05;
    const REG_CONFIGURATION: u8 = 0x06;
    const REG_WRITE_PROTECT: u8 = 0x09;
    const REG_SLEEP1: u8 = 0x0A;
    const REG_SLEEP2: u8 = 0x0B;
    const REG_SLEEP3: u8 = 0x0C;
    const REG_CPI_X: u8 = 0x0D;
    const REG_CPI_Y: u8 = 0x0E;
    const REG_DELTA_XY_HI: u8 = 0x12;
    const REG_IQC: u8 = 0x13;
    const REG_SHUTTER: u8 = 0x14;
    const REG_FRAME_AVG: u8 = 0x17;
    const REG_MOUSE_OPTION: u8 = 0x19;
    const REG_SPI_MODE: u8 = 0x26;

    pub fn new(
        pins: Pins,
        resets: &mut RESETS,
        spio: SPI0,
        peripheral_clock: impl Into<HertzU32>,
    ) -> Self {
        // These are implicitly used by the spi driver if they are in the correct mode
        let _sclk = pins.gpio2.into_mode::<gpio::FunctionSpi>();
        let _mosi = pins.gpio3.into_mode::<gpio::FunctionSpi>();
        let _miso = pins.gpio4.into_mode::<gpio::FunctionSpi>();
        let mut cs = pins.gpio5.into_push_pull_output();

        // Create an SPI driver instance for the SPI0 device
        let spi = spi::Spi::<_, _, 8>::new(spio);

        // Exchange the uninitialised SPI driver for an initialised one
        let spi = spi.init(
            resets,
            peripheral_clock,
            HertzU32::from_raw(16_000_000u32),
            &embedded_hal::spi::MODE_0,
        );

        cs.set_high().unwrap();

        Self {
            _sclk,
            _mosi,
            _miso,
            cs,

            spi,
        }
    }

    fn access<F, T>(&mut self, f: F) -> T
    where
        F: Fn(&mut Self) -> T,
    {
        self.cs.set_low().unwrap();

        let res = f(self);

        self.cs.set_high().unwrap();

        res
    }

    pub fn read(&mut self, address: u8) -> u8 {
        self.access(|me| {
            me.spi.send(address & Self::READ).unwrap();
            me.spi.read().unwrap()
        })
    }

    pub fn write(&mut self, address: u8, value: u8) {
        self.access(|me| {
            me.spi.send(address | Self::WRITE).unwrap();
            me.spi.send(value).unwrap();
        });
    }

    pub fn read_movement(&mut self) -> Option<(i8, i8)> {
        let (motion, x_overflow, y_overflow) = self.motion_detected();

        if motion {
            if x_overflow || y_overflow {
                return None;
            }

            let x = self.delta_x();
            let y = self.delta_y();

            Some((x, y))
        } else {
            None
        }
    }

    pub fn delta_x(&mut self) -> i8 {
        unsafe { mem::transmute(self.read(Self::REG_DELTA_X)) }
    }

    pub fn delta_y(&mut self) -> i8 {
        unsafe { mem::transmute(self.read(Self::REG_DELTA_Y)) }
    }

    pub fn motion_detected(&mut self) -> (bool, bool, bool) {
        let reg = self.read(Self::REG_MOTION_STATUS);

        let motion = (reg & 0b10000000) != 0;
        let x_overflow = (reg & 0b00010000) != 0;
        let y_overflow = (reg & 0b00001000) != 0;

        (motion, x_overflow, y_overflow)
    }
}
